import os
import subprocess
import time
from huepy import bad, bold

def wait_command(rg, cg, cont, wait=None):
    time_spent = 0
    exception = None
    command_uuid = None
    container_logs = subprocess.check_output(
        f'az container logs --resource-group {rg} --name {cg} --container-name {cont}', 
        shell=True
    )
    container_logs = container_logs.decode()

    while True:
        container_logs = subprocess.check_output(
        f'az container logs --resource-group {rg} --name {cg} --container-name {cont}', 
        shell=True
        )
        container_logs = container_logs.decode()

        if wait and time_spent > wait:
            exception = 'TIMEOUT REACHED'
            break
        if 'command: ' in container_logs:
            command_uuid = container_logs.split('command: ')[1].strip()
            break
        if 'Exception' in container_logs:
            exception = container_logs
            break
        time.sleep(1)
        time_spent += 1

    return cont, command_uuid, exception


def exec_command(rg, cg, cont, command, max_retries, input_file):
    env = os.environ.copy()
    env["PATH"] = "/usr/sbin:/sbin:" + env["PATH"]
    # Kill tmux window
    subprocess.Popen(["tmux", "kill-session", "-t", cont], env=env, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    time.sleep(2)
    subprocess.Popen(["tmux", "new-session", "-d", "-s", cont], env=env)
    time.sleep(5)
    subprocess.Popen(["tmux", "send-keys", "-t", cont,
                    f"az container exec -g {rg} -n {cg} --container-name {cont} --exec-command /bin/bash", "Enter",
                    ], env=env)
    time.sleep(15)
    if input_file:
        subprocess.Popen(["tmux", "send-keys", "-t", cont, f"acido -d {input_file}", "Enter"], env=env)
        time.sleep(5)
    subprocess.Popen(["tmux", "send-keys", "-t", cont, f"nohup acido -sh '{command}' > temp &", "Enter"], env=env)
    time.sleep(2)
    subprocess.Popen(["tmux", "send-keys", "-t", cont, "Enter"], env=env)

    output = subprocess.check_output(["tmux", "capture-pane", "-pt", cont], env=env).decode()

    time.sleep(4)

    retries = 0
    failed = False
    exception = None
    command_uuid = None

    while 'Done' not in output:

        retries += 1

        if retries > max_retries:
            exception = 'TIMEOUT REACHED'
            failed = True
            break

        subprocess.Popen(["tmux", "send-keys", "-t", cont, "Enter"], env=env)
        time.sleep(1)
        output = subprocess.check_output(["tmux", "capture-pane", "-pt", cont], env=env).decode()

        if 'Exit' in output:
            subprocess.Popen(["tmux", "send-keys", "-t", cont, "Enter"], env=env)
            time.sleep(2)
            subprocess.Popen(["tmux", "send-keys", "-t", cont, "cat temp", "Enter", "Enter"], env=env)
            time.sleep(2)
            subprocess.Popen(["tmux", "send-keys", "-t", cont, "Enter"], env=env)
            time.sleep(2)
            try:
                exception = subprocess.check_output(["tmux", "capture-pane", "-pt", cont], env=env)
                exception = exception.decode()
                exception = exception.split('cat temp')[1].strip()
            except Exception as e:
                exception = 'ERROR PARSING'
                print(bad(f'Error capturing output from: {bold(cont)}'))
            failed = True
            break
        if 'Done' in output:
            failed = False
            break

    if not failed:
        subprocess.Popen(["tmux", "send-keys", "-t", cont, "Enter"], env=env)
        time.sleep(2)
        subprocess.Popen(["tmux", "send-keys", "-t", cont, "cat temp", "Enter", "Enter"], env=env)
        time.sleep(10)
        subprocess.Popen(["tmux", "send-keys", "-t", cont, "Enter"], env=env)
        time.sleep(2)
        try:
            output = subprocess.check_output(["tmux", "capture-pane", "-pt", cont], env=env)
            output = output.decode()
            command_uuid = output.split('command: ')[1].split('\n')[0].strip()
        except Exception as e:
            command_uuid = None
            print(bad(f'Error capturing output from: {bold(cont)}'))
    else:
        print(bad(f'Exception ocurred while executing "{command}" from: {bold(cont)}'))

    # Kill shell
    subprocess.Popen(["tmux", "send-keys", "-t", cont, "(rm temp && exit)", "Enter"], env=env)
    time.sleep(1)
    # Kill tmux window
    subprocess.Popen(["tmux", "kill-session", "-t", cont], env=env)

    return cont, command_uuid, exception




from huepy import good, bad, info


import time
import subprocess
import os
from os.path import join as jpath
from os.path import expanduser
from base64 import b64encode, b64decode

def basic_auth_header(password):
    user_pass = f':{password}'
    basic_credentials = b64encode(user_pass.encode()).decode()
    return basic_credentials

def chunks(iterable, chunk_size):
    """Generates lists of `chunk_size` elements from `iterable`.
    >>> list(chunk((2, 3, 5, 7), 3))
    [[2, 3, 5], [7]]
    >>> list(chunk((2, 3, 5, 7), 2))
    [[2, 3], [5, 7]]
    """
    iterable = iter(iterable)
    while True:
        chunk = []
        try:
            for _ in range(chunk_size):
                chunk.append(next(iterable))
            yield chunk
        except StopIteration:
            if chunk:
                yield chunk
            break

def split_file(input_file, number_of_containers):
    number_of_lines = len([line.strip() for line in open(input_file, 'r').readlines()])
    chunked_lines = int(number_of_lines / number_of_containers)
    print(good(f'Splitting into {number_of_containers} files.'))
    os.system(f'split -l {str(chunked_lines)} {input_file} /tmp/acido-input-')

    directory = os.fsencode('/tmp/')
    input_files = []
        
    for file in os.listdir(directory):
        filename = os.fsdecode(file)
        if filename.startswith("acido-input"): 
            input_files.append(f'/tmp/{filename}')
        else:
            continue

    return input_files


from acido.azure_utils.ManagedIdentity import ManagedAuthentication, Resources
from acido.utils.decoration import __version__
from azure.mgmt.containerinstance import ContainerInstanceManagementClient
from azure.mgmt.containerinstance.models import (
    ContainerGroup, Container, ImageRegistryCredential, ResourceRequirements,
    ResourceRequests, OperatingSystemTypes, EnvironmentVariable,
    ResourceIdentityType)
from azure.mgmt.containerinstance.models import ContainerGroupIdentity, UserAssignedIdentities
from azure.core.exceptions import HttpResponseError, ResourceNotFoundError
from huepy import good, bad, info


from shlex import quote
import logging

__authors__ = "Juan Ramón Higueras Pica (juanramon.higueras@wsg127.com)"
__coauthor__ = "Xavier Álvarez Delgado (xalvarez@merabytes.com)"

logger = logging.getLogger('msrest.serialization')
logger.disabled = True

class InstanceManager(ManagedAuthentication):
    def __init__(self, resource_group, login: bool = True, user_assigned: dict = {}, network_profile=None):
        if login:
            credential = self.get_credential(Resources.INSTANCE)
            subscription = self.extract_subscription(credential)
            self._client = ContainerInstanceManagementClient(
                credential,
                subscription
            )
        self.resource_group = resource_group
        self.image_registry_credentials = None
        self.network_profile = network_profile
        self.env_vars = {}
        self.instances = []
        self.user_assigned = user_assigned

    def login_image_registry(self, server, username, password):
        image_registry_credentials = ImageRegistryCredential(
            server=server,
            username=username,
            password=password
        )
        self.image_registry_credentials = image_registry_credentials

    def provision(
        self, 
        name, 
        memory: float = 1, 
        cpu: float = 1, 
        ports=None,
        command=[
            'sleep',
            'infinity'
        ], 
        image = 'ubuntu:20.04',
        env_vars: dict = {}
    ):
        env = {}

        for env_key, env_val in env_vars.items():
            env[env_key] = EnvironmentVariable(
            name=env_key,
            value=env_val
        )
        resource_request = ResourceRequests(
            memory_in_gb=memory,
            cpu=cpu
        )
        resource_requirement = ResourceRequirements(
            requests=resource_request
        )
        instance = Container(
            name=name,
            image=image,
            resources=resource_requirement,
            command=command,
            ports=ports,
            environment_variables=list(env.values())
        )
        return instance

    def deploy(
        self, name, tags: dict = {}, location="westeurope",
        ip_address=None, os_type=OperatingSystemTypes.linux,
        restart_policy="Never", network_profile=None, 
        instance_number: int = 3, max_ram: int = 16, 
        max_cpu: int = 4, image_name: str =None,
        env_vars: dict = {}, command: str = None,
        input_files: list = None, quiet: bool = False
    ):
        restart_policies = ["Always", "OnFailure", "Never"]
        if restart_policy not in restart_policies:
            raise ValueError((
                "Unsupported restart policy \"%s\"."
                "Please use one of the following: %s"
                % (restart_policy, ", ".join(restart_policies))
            ))
        if not self.image_registry_credentials:
            ir_credentials = []
        
        if self.image_registry_credentials:
            ir_credentials = [self.image_registry_credentials]

        max_ram = "{:.1f}".format(max_ram / (instance_number + 1))
        max_cpu = "{:.1f}".format(max_cpu / (instance_number + 1)) 

        ok = False
        
        results = {}

        deploy_instances = []

        if command:
            command = f"acido -sh {quote(command)}"

        for i_num in range(1, instance_number + 1):
            env_vars['INSTANCE_NAME'] = f'{name}-{i_num:02d}'
            scan_cmd = command

            if input_files:
                file_uuid = input_files.pop(0)
                upload_command = f"acido -d {file_uuid}"
                if scan_cmd:
                    scan_cmd = upload_command + " && " + scan_cmd
                else:
                    scan_cmd = upload_command

            if scan_cmd:
                deploy_instances.append(
                    self.provision(
                        f'{name}-{i_num:02d}', 
                        memory=float(max_ram), 
                        cpu=float(max_cpu), 
                        image=image_name,
                        env_vars=env_vars,
                        command=["/bin/sh", "-c", scan_cmd]
                        )
                )
            else:
                deploy_instances.append(
                    self.provision(
                        f'{name}-{i_num:02d}', 
                        memory=float(max_ram), 
                        cpu=float(max_cpu), 
                        image=image_name,
                        env_vars=env_vars,
                        command=scan_cmd
                        )
                    )

        try:
            cg = ContainerGroup(
                location=location,
                containers=deploy_instances,
                os_type=os_type,
                ip_address=ip_address,
                image_registry_credentials=ir_credentials,  # noqa: E501
                restart_policy=restart_policy,
                tags=tags,
                network_profile=self.network_profile,
                identity=ContainerGroupIdentity(
                    type=ResourceIdentityType.user_assigned,
                    user_assigned_identities={
                        self.user_assigned.get('id', ''): UserAssignedIdentities(client_id=self.user_assigned.get('clientId', ''))
                    }
                )
            )
            self._client.container_groups.begin_create_or_update(
                resource_group_name=self.resource_group,
                container_group_name=name,
                container_group=cg
            ).result()
            ok = True
            for i_num in range(1, instance_number + 1):
                results[f'{name}-{i_num:02d}'] = ok
        except HttpResponseError as e:
            ok = False
            if not quiet:
                print(bad(str(e)))
            raise e

        self.env_vars.clear()
        return results, input_files

    def rm(self, group_name: str) -> bool:
        try:
            poller = self._client.container_groups.begin_delete(
                resource_group_name=self.resource_group,
                container_group_name=group_name,
            )
            # wait until the delete finishes (raise on failure)
            poller.result()
            return True
        except ResourceNotFoundError:
            print(f"Container group '{group_name}' not found.")
            return False
        except HttpResponseError as e:
            print(f"Delete failed: {e}")
            return False

    def get(self, group_name):
        try:
            cg = self._client.container_groups.get(
                resource_group_name=self.resource_group,
                container_group_name=group_name
            )
        except HttpResponseError as e:
            if e.status_code == 404:
                return False
            else:
                print(bad(str(e)))
        return cg

    def ls(self):
        try:
            cg = self._client.container_groups.list_by_resource_group(
                resource_group_name=self.resource_group,
            )
        except HttpResponseError as e:
            print(bad(str(e)))
            return []
        return cg

    def restart(self, group_name):
        try:
            self._client.container_groups.restart(
                resource_group_name=self.resource_group,
                container_group_name=group_name
            )
        except HttpResponseError as e:
            if e.status_code == 404:
                return False
            else:
                # We don't know what is happening, maybe transitioning?
                print(str(e))
        return True

    def stop(self, group_name):
        try:
            self._client.container_groups.stop(
                resource_group_name=self.resource_group,
                container_group_name=group_name
            )
        except HttpResponseError as e:
            print(str(e))
            return False
        return True

    def start(self, group_name):
        try:
            self._client.container_groups.start(
                resource_group_name=self.resource_group,
                container_group_name=group_name
            )
        except HttpResponseError as e:
            if e.status_code == 404:
                return False
            else:
                # We don't know what is happening, maybe transitioning?
                print(str(e))
        return True
